/*
 * This was originally written by Mark Hahn.  Obtained from
 * http://brain.mcmaster.ca/~hahn/realfeel.c
 *
 * Modified by Esben Nielsen
 */

#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sched.h>
#include <sys/signal.h>
#include <string.h>

#include <pthread.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "rdtsc.h"
#include "blocker.h"

/* global vars */
int stopit;				/* set to stop measuring */

#define SAMPLES	10000
int histogram1[SAMPLES];			/* Milliseconds */

#define PAGE_SIZE	4096UL		/* virtual memory page size */
#define OSCR_HZ		3686400		/* frequency of clock ticks */
#define OSCR		0x90000010	/* physical address of OSCR register */
unsigned long *oscr;			/* ptr to OSCR */


void setup_clock (void)
{
	int fd;
	void *map_base;
	off_t target;
	off_t page;

#ifndef ARCHARM
	return;
#endif

	fd = open ("/dev/mem", O_RDWR | O_SYNC);
	if (-1 == fd) {
	    perror ("open of /dev/mem failed\n");
	    exit (1);
	}

	/* Map one page */
	target = OSCR;
	page = target & ~(PAGE_SIZE - 1);
	map_base = mmap (0, PAGE_SIZE, PROT_READ, MAP_SHARED, fd, page);
	if (MAP_FAILED == map_base) {
	    perror ("mmap failed");
	    (void) close (fd);
	    exit (2);
	}
	oscr = map_base + (target - page);

	/* This does not end the mmap,
	 * the mmap will go away when the process dies
	 */
	close (fd);
}

double second() {
	struct timeval tv;
	gettimeofday(&tv,0);
	return tv.tv_sec + 1e-6 * tv.tv_usec;
}


void selectsleep(unsigned us) {
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = us;
	select(0,0,0,0,&tv);
}

double secondsPerTick, ticksPerSecond;

void calibrate()
{
	double sumx = 0;
	double sumy = 0;
	double sumxx = 0;
	double sumxy = 0;
	double slope;

	// least squares linear regression of ticks onto real time
	// as returned by gettimeofday.

	const unsigned n = 30;
	unsigned i;

	for (i=0; i<n; i++) {
		double breal,real,ticks;
		u64 bticks;
	
		breal = second();
		bticks = get_cpu_tick();

		selectsleep((unsigned)(10000 + drand48() * 200000));

		ticks = get_cpu_tick() - bticks;
		real = second() - breal;

		sumx += real;
		sumxx += real * real;
		sumxy += real * ticks;
		sumy += ticks;
	}
	slope = ((sumxy - (sumx*sumy) / n) /
		 (sumxx - (sumx*sumx) / n));
	ticksPerSecond = slope;
	secondsPerTick = 1.0 / slope;
	printf("%3.3f MHz\n",ticksPerSecond*1e-6);
}

void fatal(char *msg)
{
	perror(msg);
	exit(1);
}

int set_realtime_priority(void)
{
	struct sched_param schp;
	/*
	 * set the process to realtime privs
	 */
	memset(&schp, 0, sizeof(schp));
	schp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	
	printf("Priority: %d\n",schp.sched_priority);
	if (sched_setscheduler(0, SCHED_FIFO, &schp) != 0) {
		perror("sched_setscheduler");
		exit(1);
	}

	return 0;
}

void hist(char *file)
{
	int i;
	FILE *f;

	f = fopen(file, "w");
	if (f == 0) {
		fprintf(stderr, "realfeel: can't open `%s':%s\n",
			file, strerror(errno));
		exit(1);
	}

	for (i = 0; i < SAMPLES; i++) {
		if (histogram1[i]) {
		  fprintf(f, "%d.%d %d\n",
			  i / 10, i % 10, histogram1[i]);
		}
	}
	fclose(f);
	exit(0);
}

void signalled(int sig)
{
	stopit = 1;
}

static void usage(const char *arg0)
{
	fprintf(stderr, "Usage: %s [--samples n] [--hertz n] [--tasks n] "
		"filename.hist\n",
		arg0);
	exit(1);
}

uint bounded=0;
uint ncycles=0;
uint current=0;
int hz = 128;
int ntasks = 1;

char *parse_options(int argc, char **argv)
{
  int c, option_index;

  static struct option long_options[] = {
    {"samples", required_argument, 0, 0},
    {"hertz", required_argument, 0, 0}, 
    {"tasks", required_argument, 0, 0},
    {0, 0, 0, 0}, 
  };
  
  while ((c = getopt_long (argc, argv, "b:", long_options, &option_index)) != -1) {
    switch (option_index) {
      
    default:
    case -1:
      fprintf (stderr, "Bad option (%s)\n", optarg);
      usage(argv[0]);
      exit (1);
      break;

    case 0:
      bounded = 1;
      ncycles = atoi(optarg);
      break;

    case 1:
      hz = atoi(optarg);
      if (hz > 128) {
	fprintf(stderr, "max allowable interrupt frequency is 2048 Hz\n");
	hz = 128;
      }
      else if (hz <= 0) {
	fprintf(stderr, "zero or negative frequency doesn't make sense!\n");
	hz = 1;
      }
      break;
    case 2:
      ntasks = atoi(optarg);
      if(ntasks<0) {
	fprintf(stderr, "negative number of tasks make no sense!\n");
	ntasks = 0;
      }
      else if(ntasks>20) {
	fprintf(stderr, "Now relax a bit!\n");
	ntasks = 20;
      }
      break;
    }
  }

  if (argv[optind] == NULL) {
    fprintf(stderr, "histogram file name required\n");
    usage(argv[0]);
    exit (2);
  }

  return argv[optind];
}

int loopsPer1ms;

void loop(int loops)
{
  int i;
  for(i=0;i<loops; i++)
    get_cpu_tick(); /* Just to do something which can't be optimized out! */
}

void calculate_loops()
{
  int i;
  loopsPer1ms = 10000;
  for(i=0;i<10;i++) {
    u64 delay,begin = get_cpu_tick();
    loop(loopsPer1ms);
    delay = get_cpu_tick()-begin;
    printf("loops: %d, delay: %g\n",loopsPer1ms, (1.0*delay)/ticksPerSecond);
    loopsPer1ms = (loopsPer1ms*i + 
		   2*(loopsPer1ms*ticksPerSecond)/(1000*delay))/ (i+2);
  }
    

  printf("Loops per 1ms: %d\n",loopsPer1ms);
    
}

#define TEST_PTHREAD_MUTEX
#ifdef TEST_PTHREAD_MUTEX
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
#else
int blocker_fd;
#endif

void lock_and_spin(int loops)
{
#ifdef TEST_PTHREAD_MUTEX
      pthread_mutex_lock(&mutex);
      loop(loops);
      pthread_mutex_unlock(&mutex);
#else
      ioctl(blocker_fd,BLOCK_IOCTL,loops);
#endif
}

void *dummy_loop(void *arg)
{
  while(!stopit)
    {
      lock_and_spin(loopsPer1ms);
    }

  return arg;
}

void *dummy_loop_empty(void *arg)
{
  while(!stopit)
    {
    }

  return arg;
}

void spawn_task(int i)
{
  pthread_t th;
  pthread_attr_t attr;
  
  if(pthread_attr_init(&attr))
    {
      perror("pthread_attr_init:");
      exit(1);
    }
  
  pthread_attr_setschedpolicy(&attr,SCHED_OTHER);
  
  if(pthread_create(&th, &attr, i==-1?dummy_loop_empty:dummy_loop, NULL))
    {
      perror("pthread_create:");
      exit(1);
    }
}

#define msec(f) (1e3 * (f))

int main(int argc, char *argv[])
{
	int fd;
	u64 last;
	double max_delay = 0;
	char *histfile = parse_options(argc, argv);

	blocker_fd = open("/dev/blocker",O_RDONLY);
	if (blocker_fd<0) {
	  perror("open /dev/blocker");
	  exit(1);
	}
	  

	if (mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
		perror("mlockall");
		exit(1);
	}
	
	setup_clock ();

	{
	  int i;
	  for(i=-1;i<ntasks;i++) {	  
	    spawn_task(i);
	  }
	}

 	set_realtime_priority();
	calibrate();
	printf("secondsPerTick=%f\n", secondsPerTick);
	printf("ticksPerSecond=%f\n", ticksPerSecond);

	calculate_loops();

	if (bounded) {
	  printf("running for %d samples\n", ncycles);
	}

	signal(SIGINT, signalled);

	while (!stopit) {
		u64 now1,now2;
		double delay;
		int data;
		int ms;

		selectsleep(100000); /* 100 ms */
		now1 = get_cpu_tick();
		lock_and_spin(0);
		now2 = get_cpu_tick();

		delay = secondsPerTick * (now2 - now1);
		if (delay > max_delay) {
			max_delay = delay;
		//	printf("%.3f msec\n", (1e3 * (delay)));
		}

		ms = (delay + 1.0/20000.0) * 10000;
		if (ms < 0)
			ms = 0;		/* hmmm */
		if (ms >= SAMPLES)
			ms = SAMPLES;
		histogram1[ms]++;

		if (bounded) {
		  if (++current >= ncycles) {
		    printf ("finished collecting %d samples\n", ncycles);
		    printf ("maximum cycle time: %.3fms\n", 
			    msec(max_delay));
		    break;
		  }

		  if ((current % 10000) == 0) {
		    printf("%d cycles (max cycle time so far: %.3fms)\n",
			   current, -(msec(max_delay)));
		  }
		}

	}
	hist(histfile);

	return 0;
}
